/*
 * Copyright (c) Members of the EGEE Collaboration. 2004. 
 * See http://www.eu-egee.org/partners/ for details on the copyright
 * holders.  
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 *
 *     http://www.apache.org/licenses/LICENSE-2.0 
 *
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License.
 */
 
/*
 *
 * Author Luigi Zangrando <zangrando@pd.infn.it>
 *
 */

package org.glite.ce.monitorapij.resource;

import java.io.File;
import java.net.URI;
import java.net.URL;
import java.util.ArrayList;
import java.util.Hashtable;
import java.util.List;

import javax.naming.Context;
import javax.naming.NamingEnumeration;
import javax.naming.NamingException;
import javax.naming.directory.Attribute;
import javax.naming.directory.Attributes;
import javax.naming.directory.BasicAttributes;
import javax.naming.directory.DirContext;
import javax.naming.directory.InitialDirContext;
import javax.naming.directory.SearchResult;

import org.apache.log4j.Logger;
import org.glite.ce.commonj.utils.CEUtils;
import org.glite.ce.commonj.utils.JarClassLoader;
import org.glite.ce.commonj.utils.TimerTask;
import org.glite.ce.monitorapij.resource.types.CEMonResource;
import org.glite.ce.monitorapij.resource.types.Property;

/**
 * A class used to handle a list of <code>Resource</code>s placed in a specified
 * directory as plugins. An XML config file is used to store and retrieve
 * information about these resources. The task of checking if resources are
 * added or removed or modified can be scheduled for repeated executions.
 */
public abstract class ResourceHolder extends TimerTask {
    private final static Logger logger = Logger.getLogger(ResourceHolder.class.getName());
    private ArrayList<ResourceListener> listeners;
    private Hashtable<String, Resource> resourceList;
    private long lastConfigFileModified;

    protected Hashtable<String, String> jndiProperties;
    private boolean exit = false;

    /**
     * Creates a new ResourceHolder object specifying the current
     * <code>MessageContext</code>, the resource path and the config file path.
     * 
     * @param context
     *            The current <code>MessageContext</code>.
     * @param resourcePath
     *            The path where new resources are placed.
     * @param configFilePath
     *            The path of the config file.
     */
    public ResourceHolder() {
        resourceList = new Hashtable<String, Resource>(0);
        listeners = new ArrayList<ResourceListener>(0);
        lastConfigFileModified = 0;

        jndiProperties = null;
        String confProvider = CEUtils.getConfigurationProvider(CEUtils.CEMONITOR_LABEL);
        String confProviderURL = CEUtils.getConfigurationURL(CEUtils.CEMONITOR_LABEL);

        if (confProvider != null && confProviderURL != null) {
            jndiProperties = new Hashtable<String, String>(2);
            jndiProperties.put(Context.INITIAL_CONTEXT_FACTORY, confProvider);
            jndiProperties.put(Context.PROVIDER_URL, confProviderURL);
        }
    }

    /**
     * Add a <code>Resource</code> to the list of handled resources.
     * 
     * @param resource
     *            The <code>Resource</code> to add. This must have a defined ID.
     */
    public void addResource(Resource resource) {
        if (resource != null) {
            if (getResource(resource.getId()) == null) {
                resourceList.put(resource.getId(), resource);
            }
        }
    }

    /**
     * Remove the specified <code>Resource</code>.
     * 
     * @param resource
     *            The resource to be removed.
     */
    public void removeResource(Resource resource) {
        if (resource != null) {
            if (getResource(resource.getId()) != null) {
                resourceList.remove(resource.getId());
            }
        }
    }

    /**
     * Get the <code>Resource</code> specified by its ID.
     * 
     * @param id
     *            The ID int of the <code>Resource</code> searched.
     * @return The Resource searched if present in the list of handled
     *         resources, null otherwise.
     */
    public Resource getResource(String id) {
        Resource resource = null;

        if (id != null) {
            resource = (Resource) resourceList.get(id);
        }

        return resource;
    }

    public int getResourceListSize() {
        return resourceList.size();
    }

    /**
     * Get all the handled resources.
     * 
     * @return An array containing all the handled resources.
     */
    public Resource[] getResources() {
        Resource[] resources = new Resource[resourceList.values().size()];
        resources = (Resource[]) resourceList.values().toArray(resources);
        return resources;
    }

    public List<Resource> getResourceList() {
        return new ArrayList<Resource>(resourceList.values());
    }

    /**
     * Get the all <code>Resource</code> of the specified type.
     * 
     * @param id
     *            The type of the <code>Resource</code>s searched.
     * @return An Array containing the Resources searched if present in the list
     *         of handled resources, an empty array otherwise.
     */
    public Resource[] getResourceByType(String type) {
        ArrayList<Resource> tmp = new ArrayList<Resource>(0);
        Resource[] resourceArray = getResources();

        for (int i = 0; i < resourceArray.length; i++) {
            if (resourceArray[i].getType().equals(type)) {
                tmp.add(resourceArray[i]);
            }
        }

        return makeResources(tmp);
    }

    /**
     * Get the <code>CEMonResource</code>s with the specified jar path.
     * 
     * @param path
     *            The path of the jar of the resource.
     * @return An array containing the searched resources if present in the list
     *         of handled resources.
     */
    public Resource[] getResourceByJARPath(URI path) {
        ArrayList<Resource> tmp = new ArrayList<Resource>(0);
        Resource[] resourceArray = getResources();

        for (int i = 0; i < resourceArray.length; i++) {
            if (resourceArray[i].getJarPath().equals(path)) {
                tmp.add(resourceArray[i]);
            }
        }

        return makeResources(tmp);
    }

    /**
     * Get the <code>CEMonResource</code>s with the specified name.
     * 
     * @param name
     *            The name of the searched resource.
     * @return An array containing the searched resources if present in the list
     *         of handled resources.
     */
    public Resource[] getResourceByName(String name) {
        ArrayList<Resource> tmp = new ArrayList<Resource>(0);
        Resource[] resourceArray = getResources();

        for (int i = 0; i < resourceArray.length; i++) {
            if (resourceArray[i].getName().equals(name)) {
                tmp.add(resourceArray[i]);
            }
        }

        return makeResources(tmp);
    }

    /**
     * Convert the specified ArrayList in an array of Resources.
     * 
     * @param list
     *            The specified ArrayList
     * @return The Array of Resources containing all the handled Resources.
     */
    private Resource[] makeResources(ArrayList<Resource> list) {
        Resource[] resources = new Resource[list.size()];
        resources = (Resource[]) list.toArray(resources);

        return resources;
    }

    /**
     * Add an array of <code>Resource</code>s to the list of handled resources.
     * 
     * @param resources
     *            An array containing the resources to add.
     */
    public void setResources(Resource[] resources) {
        clear();

        if (resources == null) {
            return;
        }

        for (int i = 0; i < resources.length; i++) {
            addResource(resources[i]);
        }
    }

    /**
     * Clear the list of handled resources.
     */
    public void clear() {
        resourceList.clear();
    }

    /**
     * Check if the <code>Property</code> array of the two specified
     * <code>Resource</code>s equals.
     * 
     * @param resource
     *            The first <code>Resource</code> to analyze.
     * @param new_resource
     *            The second <code>Resource</code> to analyze.
     * @return true if properties are the same, false otherwise.
     */
    private boolean checkEqualProperty(Resource resource, Resource new_resource) {
        Property[] property = resource.getProperty();
        Property[] new_property = new_resource.getProperty();

        if ((property == null) && (new_property == null)) {
            return true;
        }

        if ((property == null) ^ (new_property == null)) {
            return false;
        }

        if (property.length != new_property.length) {
            return false;
        }

        for (int i = 0; i < property.length; i++) {
            if (new_resource.getProperty(property[i].getName()) == null) {
                return false;
            }
        }

        return true;
    }

    /**
     * DOCUMENT ME!
     * 
     * @param resource
     *            DOCUMENT ME!
     * @return DOCUMENT ME!
     * @throws Exception
     *             DOCUMENT ME!
     */
    private Resource makeResourceInstance(Resource resource) throws Exception {
        if (!resource.getClass().equals(CEMonResource.class)) {
            return resource;
        }

        Resource newResource = makeResourceInstance(resource.getJarPath());
        newResource.setId(resource.getId());
        newResource.setJarPath(resource.getJarPath());
        newResource.setCreationTime(resource.lastModified());

        Property[] allProps = newResource.getProperty();
        Property[] props = resource.getProperty();

        Hashtable<String, Property> propertyTable = new Hashtable<String, Property>();

        if (props != null) {
            for (int i = 0; i < props.length; i++) {
                logger.debug("makeResourceInstance - found new resource property " + props[i].getName());
                propertyTable.put(props[i].getName(), props[i]);
            }
        }

        if (allProps != null) {
            for (int i = 0; i < allProps.length; i++) {
                if (!propertyTable.containsKey(allProps[i].getName())) {
                    logger.debug("makeResourceInstance - found resource property " + allProps[i].getName());
                    propertyTable.put(allProps[i].getName(), allProps[i]);
                }
            }
        }

        Property[] propertyMerged = new Property[propertyTable.size()];
        propertyMerged = (Property[]) propertyTable.values().toArray(propertyMerged);

        newResource.setProperty(propertyMerged);

        return newResource;
    }

    /**
     * DOCUMENT ME!
     * 
     * @param jarFile
     *            DOCUMENT ME!
     * @return DOCUMENT ME!
     * @throws Exception
     *             DOCUMENT ME!
     */
    private Resource makeResourceInstance(URI jarFile) throws Exception {
        return makeResourceInstance(new File(jarFile.getPath()));
    }

    /**
     * DOCUMENT ME!
     * 
     * @param resourceJarFile
     *            DOCUMENT ME!
     * @return DOCUMENT ME!
     * @throws Exception
     *             DOCUMENT ME!
     */
    protected Resource makeResourceInstance(File resourceJarFile) throws Exception {
        Class c = makeResourceClass(resourceJarFile);

        if (c != null) {
            Object obj = c.newInstance();

            if (obj instanceof Resource) {
                Resource resource = (Resource) obj;
                resource.setJarPath(new URI("file://:" + resourceJarFile.getAbsolutePath()));

                return resource;
            }
        }

        throw (new Exception("Cannot create a Resource instance from file \"" + resourceJarFile.getAbsolutePath() + "\""));
    }

    /**
     * Get an instance of the <code>Resource</code> class packed in the jar
     * file.
     * 
     * @param resourceJarFile
     *            The jar file containing the class of a <code>Resource</code>.
     *            It must contain the "Main-Class" attribute into its
     *            MANIFEST.MF specifying the name of the Resource class.
     * @return The class contained in the jar file.
     * @throws Exception
     *             If the jar file does not contain a main class the
     *             "Main-Class" attribute into its MANIFEST.MF or if the
     *             instance cannot be created.
     */
    protected Class makeResourceClass(File resourceJarFile) throws Exception {
        if (resourceJarFile == null) {
            throw (new Exception("The jar file is null!"));
        }

        if (resourceJarFile.getName().endsWith(".jar")) {
            URL sensorURL = resourceJarFile.toURI().toURL();

            JarClassLoader loader = new JarClassLoader("file://:" + resourceJarFile.getPath(), this.getClass().getClassLoader());
            loader.addJarURL(sensorURL);

            String className = loader.getMainClassName(sensorURL);

            if (className != null) {
                Class c = loader.loadClass(className);

                return c;
            } else {
                throw (new Exception("\"Main-Class\" attribute not found into the MANIFEST.MF of " + resourceJarFile.getAbsolutePath() + ".jar"));
            }
        }

        throw (new Exception("Cannot create a Resource instance from file \"" + resourceJarFile.getAbsolutePath() + "\""));
    }

    /**
     * DOCUMENT ME!
     * 
     * @param eventID
     *            DOCUMENT ME!
     * @param resourceList
     *            DOCUMENT ME!
     */
    protected synchronized void fireEvent(int eventID, ArrayList<Resource> resourceList) {
        logger.debug("ResourceHolder  fireEvent " + resourceList.size() + " eventId=" + eventID);

        if ((resourceList != null) && (resourceList.size() > 0)) {
            ResourceEvent event = new ResourceEvent(eventID, resourceList, this);

            // Process the listeners first to last, notifying
            // those that are interested in this event
            for (int i = 0; i < listeners.size(); i++) {
                Object listener = listeners.get(i);

                if (listener instanceof ResourceListener) {
                    ((ResourceListener) listener).doOnResourceEvent(event);
                }
            }
        }
    }

    /**
     * This is the scheduled method checking if new resources are added or if
     * already present resources has been updated. If this is the case handles
     * the new or updated resources.
     */
    public void run() {
        logger.debug("Begin run method for " + getCategory());

        if (jndiProperties != null && !exit) {
            Hashtable<String, Resource> updatedList = new Hashtable<String, Resource>(0);
            Hashtable<String, Resource> unchangedList = new Hashtable<String, Resource>(0);
            Hashtable<String, Resource> addedList = new Hashtable<String, Resource>(0);
            DirContext rootCtx = null;

            try {
                logger.debug("Query for category " + getCategory());
                rootCtx = (DirContext) new InitialDirContext(jndiProperties);

                Attribute lmtAttr = rootCtx.getAttributes("").get("lastModificationTime");
                if (lmtAttr != null) {
                    long lmt = ((Long) lmtAttr.get(0)).longValue();

                    if (lastConfigFileModified >= lmt) {
                        logger.debug("Configuration files unchanged");
                        return;
                    }

                    lastConfigFileModified = lmt;
                    updatedList = new Hashtable<String, Resource>(0);
                    unchangedList = new Hashtable<String, Resource>(0);
                    addedList = new Hashtable<String, Resource>(0);
                }

                BasicAttributes queryAttrs = new BasicAttributes("category", getCategory());
                NamingEnumeration<SearchResult> queryResult = rootCtx.search("", queryAttrs);

                while (queryResult.hasMoreElements()) {
                    SearchResult resultItem = queryResult.next();
                    Attributes attributes = resultItem.getAttributes();
                    Resource resource = (Resource) resultItem.getObject();
                    logger.debug("Found " + resultItem.getName());

                    long scanTime = ((Long) attributes.get("scantime").get(0)).longValue();
                    long crTime = ((Long) attributes.get("creationtime").get(0)).longValue();
                    long modTime = ((Long) attributes.get("modificationtime").get(0)).longValue();

                    try {
                        if (crTime >= scanTime) {
                            addedList.put(resource.getId(), makeResourceInstance(resource));
                            logger.debug("New resource " + resource.getName());
                        } else if (modTime >= scanTime) {
                            updatedList.put(resource.getId(), makeResourceInstance(resource));
                            logger.debug("Updated resource " + resource.getName());
                        } else {
                            Resource oldRes = getResource(resource.getId());
                            if (oldRes != null) {
                                unchangedList.put(resource.getId(), oldRes);
                                logger.debug("Unchanged resource " + resource.getName());
                            }
                        }
                    } catch (Exception ex) {
                        logger.error(ex.getMessage(), ex);
                    }
                }

                if (updatedList.size() == 0 && addedList.size() == 0 && unchangedList.size() == resourceList.size()) {
                    logger.debug("Configuration unchanged");
                } else {
                    ArrayList<Resource> updatedResourcelist = new ArrayList<Resource>(0);

                    for (String resId : updatedList.keySet()) {
                        if (resourceList.containsKey(resId)) {
                            updatedResourcelist.add(updatedList.get(resId));
                        }
                    }

                    unchangedList.putAll(updatedList);

                    ArrayList<String> removedIdSet = new ArrayList<String>(resourceList.keySet());
                    ArrayList<String> unchangedIdSet = new ArrayList<String>(unchangedList.keySet());
                    removedIdSet.removeAll(unchangedIdSet);

                    ArrayList<Resource> removedResourcelist = new ArrayList<Resource>(removedIdSet.size());
                    for (String id : removedIdSet) {
                        removedResourcelist.add(resourceList.remove(id));
                    }

                    unchangedList.putAll(addedList);
                    resourceList.clear();
                    resourceList.putAll(unchangedList);

                    fireEvent(ResourceEvent.RESOURCE_REMOVED, removedResourcelist);
                    fireEvent(ResourceEvent.RESOURCE_ADDED, new ArrayList<Resource>(addedList.values()));
                    fireEvent(ResourceEvent.RESOURCE_UPDATED, updatedResourcelist);
                }
            } catch (NamingException nEx) {
                nEx.printStackTrace();
                logger.error(nEx.getMessage());
            } catch (Exception e) {
                e.printStackTrace();
            } finally {
                if (rootCtx != null) {
                    try {
                        rootCtx.close();
                    } catch (Exception ex) {
                    }
                }
            }
        }
        logger.debug("End run method for " + getCategory());
    }

    /**
     * Returns an array of all the <code>SensorListener</code> s added to this
     * holder with addSensorListener().
     * 
     * @return all of the <code>SensorListener</code> s added or an empty array
     *         if no listeners have been added
     */
    public ResourceListener[] getSensorListeners() {
        ResourceListener[] resourceListener = new ResourceListener[listeners.size()];

        return (ResourceListener[]) listeners.toArray(resourceListener);
    }

    /**
     * Adds a <code>ResourceListener</code> to the holder.
     * 
     * @param l
     *            the <code>ResourceListener</code> to be added
     */
    public void addResourceListener(ResourceListener l) {
        if (l != null) {
            listeners.add(l);
        }
    }

    /**
     * Removes a <code>ResourceListener</code> from the holder.
     * 
     * @param l
     *            the listener to be removed
     */
    public void removeResourceListener(ResourceListener l) {
        if (l != null) {
            listeners.remove(l);
        }
    }

    protected ArrayList listDiff(ArrayList l1, ArrayList l2) {
        ArrayList list1 = (l1.size() > l2.size()) ? l1 : l2;
        ArrayList list2 = (l1.size() > l2.size()) ? l2 : l1;

        ArrayList result = new ArrayList(list1.size() - list2.size());

        for (int k = 0; k < list1.size(); k++) {
            CEMonResource resource = (CEMonResource) list1.get(k);
            int idx = -1;
            for (int r = 0; r < list2.size(); r++) {
                CEMonResource tmpr = (CEMonResource) list2.get(r);
                if (resource.getName().equals(tmpr.getName())) {
                    idx = r;
                    break;
                }
            }

            if (idx < 0)
                result.add(resource);
        }

        return result;
    }

    public boolean cancel() {
        exit = true;
        listeners.clear();

        return super.cancel();
    }

    public void destroy() {
        clear();
        cancel();
    }

    public abstract String getCategory();    
}
